@@ -1,7 +1,7 @@
 #' covr: Test coverage for packages
 #'
 #' covr tracks and reports code coverage for your package and (optionally)
-#' upload the results to a coverage service like 'Codecov' <https://codecov.io> or
+#' upload the results to a coverage service like 'Codecov' <https://about.codecov.io> or
 #' 'Coveralls' <https://coveralls.io>. Code coverage is a measure of the amount of
 #' code being exercised by a set of tests. It is an indirect measure of test
 #' quality and completeness. This package is compatible with any testing
@@ -15,6 +15,66 @@
 #' # If run with no arguments `report()` implicitly calls `package_coverage()`
 #' report()
 #' ```
+#'
+#' @section Package options:
+#'
+#' `covr` uses the following [options()] to configure behaviour:
+#'
+#' \itemize{
+#'   \item `covr.covrignore`: A filename to use as an ignore file,
+#'     listing glob-style wildcarded paths of files to ignore for coverage
+#'     calculations. Defaults to the value of environment variable
+#'     `COVR_COVRIGNORE`, or `".covrignore"`  if the neither the option nor the
+#'     environment variable are set.
+#'
+#'   \item `covr.exclude_end`: Used along with `covr.exclude_start`, an optional
+#'     regular expression which ends a line-exclusion region. For more
+#'     details, see `?exclusions`.
+#'
+#'   \item `covr.exclude_pattern`: An optional line-exclusion pattern. Lines
+#'     which match the pattern will be excluded from coverage. For more details,
+#'     see `?exclusions`.
+#'
+#'   \item `covr.exclude_start`: Used along with `covr.exclude_end`, an optional
+#'     regular expression which starts a line-exclusion region. For more
+#'     details, see `?exclusions`.
+#'
+#'   \item `covr.filter_non_package`: If `TRUE` (the default behavior), coverage
+#'     of files outside the target package are filtered from coverage output.
+#'
+#'   \item `covr.fix_parallel_mcexit`:
+#'
+#'   \item `covr.flags`:
+#'
+#'   \item `covr.gcov`: If the appropriate gcov version is not on your path you
+#'     can use this option to set the appropriate location. If set to "" it will
+#'     turn off coverage of compiled code.
+#'
+#'   \item `covr.gcov_additional_paths`:
+#'
+#'   \item `covr.gcov_args`:
+#'
+#'   \item `covr.icov`:
+#'
+#'   \item `covr.icov_args`:
+#'
+#'   \item `covr.icov_flags`:
+#'
+#'   \item `covr.icov_prof`:
+#'
+#'   \item `covr.rstudio_source_markers`: A logical value. If `TRUE` (the
+#'     default behavior), source markers are displayed within the RStudio IDE
+#'     when using `zero_coverage`.
+#'
+#'   \item `covr.record_tests`: If `TRUE` (default `NULL`), record a listing of
+#'     top level test expressions and associate tests with `covr` traces
+#'     evaluated during the test's execution. For more details, see
+#'     `?covr.record_tests`.
+#'
+#'   \item `covr.showCfunctions`:
+#' }
+#'
+#'
 "_PACKAGE"
 
 #' @import methods
@@ -47,12 +107,55 @@ save_trace <- function(directory) {
   saveRDS(.counters, file = tmp_file)
 }
 
+#' Convert a counters object to a coverage object
+#'
+#' @param counters An environment of covr trace results to convert to a coverage
+#'   object. If `counters` is not provided, the `covr` namespace value
+#'   `.counters` is used.
+#' @param ... Additional attributes to include with the coverage object.
+#'
+as_coverage <- function(counters = NULL, ...) {
+  if (missing(counters))
+    counters <- .counters
+
+  counters <- as.list(counters)
+  counters <- as_coverage_with_tests(counters)
+
+  structure(counters, ..., class = "coverage")
+}
+
+#' Clean and restructure counter tests for a coverage object
+#'
+#' For tests produced with `options(covr.record_tests)`, prune any unused
+#' records in the $tests$tally matrices of each trace and get rid of the
+#' wrapping $tests environment (reassigning with value of $tests$tally)
+#'
+#' @inheritParams as_coverage
+#'
+as_coverage_with_tests <- function(counters) {
+  clean_coverage_tests(counters)
+
+  # unnest environment-wrapped $tests$tally as more accessible $tests
+  for (i in seq_along(counters)) {
+    if (!is.environment(counters[[i]]$tests)) next
+    counters[[i]]$tests <- counters[[i]]$tests$tally
+  }
+
+  tests <- counters$tests
+  counters$tests <- NULL
+  structure(counters, tests = tests, class = "coverage")
+}
+
 #' Calculate test coverage for a specific function.
 #'
 #' @param fun name of the function.
 #' @param code expressions to run.
 #' @param env environment the function is defined in.
 #' @param enc the enclosing environment which to run the expressions.
+#' @examples
+#' add <- function(x, y) { x + y }
+#' function_coverage(fun = add, code = NULL) # 0% coverage
+#' function_coverage(fun = add, code = add(1, 2) == 3) # 100% coverage
 #' @export
 function_coverage <- function(fun, code = NULL, env = NULL, enc = parent.frame()) {
   if (is.function(fun)) {
@@ -81,7 +184,7 @@ function_coverage <- function(fun, code = NULL, env = NULL, enc = parent.frame()
     eval(code, enc)
   )
 
-  structure(as.list(.counters), class = "coverage")
+  as_coverage(as.list(.counters))
 }
 
 #' Calculate test coverage for sets of files
@@ -95,6 +198,16 @@ function_coverage <- function(fun, code = NULL, env = NULL, enc = parent.frame()
 #'   functions
 #' @param parent_env The parent environment to use when sourcing the files.
 #' @inheritParams package_coverage
+#' @examples
+#' # For the purpose of this example, save code containing code and tests to files
+#' cat("add <- function(x, y) { x + y }", file="add.R")
+#' cat("add(1, 2) == 3", file="add_test.R")
+#'
+#' # Use file_coverage() to calculate test coverage
+#' file_coverage(source_files = "add.R", test_files = "add_test.R")
+#'
+#' # cleanup
+#' file.remove(c("add.R", "add_test.R"))
 #' @export
 file_coverage <- function(
   source_files,
@@ -121,7 +234,7 @@ file_coverage <- function(
       sys.source, keep.source = TRUE, envir = env)
   )
 
-  coverage <- structure(as.list(.counters), class = "coverage")
+  coverage <- as_coverage(.counters)
 
   exclude(coverage,
     line_exclusions = line_exclusions,
@@ -138,6 +251,10 @@ file_coverage <- function(
 #' @param test_code A character vector of test code
 #' @inheritParams file_coverage
 #' @param ... Additional arguments passed to [file_coverage()]
+#' @examples
+#' source <- "add <- function(x, y) { x + y }"
+#' test <- "add(1, 2) == 3"
+#' code_coverage(source, test)
 #' @export
 code_coverage <- function(
    source_code,
@@ -178,7 +295,7 @@ environment_coverage <- function(
       sys.source, keep.source = TRUE, envir = exec_env)
   )
 
-  coverage <- structure(as.list(.counters), class = "coverage")
+  coverage <- as_coverage(.counters)
 
   exclude(coverage,
     line_exclusions = line_exclusions,
@@ -214,7 +331,8 @@ environment_coverage <- function(
 #' is simply summed into one coverage object. If `FALSE` separate objects
 #' are used for each type of coverage.
 #' @param relative_path whether to output the paths as relative or absolute
-#' paths.
+#'   paths. If a string, it is interpreted as a root path and all paths will be
+#'   relative to that root.
 #' @param quiet whether to load and compile the package quietly, useful for
 #' debugging errors.
 #' @param clean whether to clean temporary output files after running, mainly
@@ -227,6 +345,10 @@ environment_coverage <- function(
 #' @param ... Additional arguments passed to [tools::testInstalledPackage()].
 #' @param exclusions \sQuote{Deprecated}, please use \sQuote{line_exclusions} instead.
 #' @param pre_clean whether to delete all objects present in the src directory before recompiling
+#' @param install_path The path the instrumented package will be installed to
+#'   and tests run in. By default it is a path in the R sessions temporary
+#'   directory. It can sometimes be useful to set this (along with `clean =
+#'   FALSE`) to help debug test failures.
 #' @seealso [exclusions()] For details on excluding parts of the
 #' package from the coverage calculations.
 #' @export
@@ -239,14 +361,16 @@ package_coverage <- function(path = ".",
                              line_exclusions = NULL,
                              function_exclusions = NULL,
                              code = character(),
+                             install_path = temp_file("R_LIBS"),
                              ...,
                              exclusions, pre_clean=TRUE) {
 
   if (!missing(exclusions)) {
-    warning(paste0("`exclusions` is deprecated and will be removed in an upcoming
-      release. ", "Please use `line_exclusions` instead."), call. = FALSE,
-      domain = NA)
-      line_exclusions <- exclusions
+    warning(
+      "`exclusions` is deprecated and will be removed in an upcoming release. Please use `line_exclusions` instead.",
+      call. = FALSE, domain = NA
+    )
+    line_exclusions <- exclusions
   }
 
   pkg <- as_package(path)
@@ -275,8 +399,16 @@ package_coverage <- function(path = ".",
     return(res)
   }
 
-  tmp_lib <- temp_file("R_LIBS")
-  dir.create(tmp_lib)
+  if (is.character(relative_path)) {
+    stopifnot(length(relative_path) == 1)
+    root <- normalize_path(relative_path)
+  } else if (isTRUE(relative_path)) {
+    root <- pkg$path
+  } else {
+    root <- NULL
+  }
+
+  dir.create(install_path)
 
   flags <- getOption("covr.flags")
 
@@ -300,7 +432,7 @@ package_coverage <- function(path = ".",
       clean_objects(pkg$path)
       clean_gcov(pkg$path)
       clean_parse_data()
-      unlink(tmp_lib, recursive = TRUE)
+      unlink(install_path, recursive = TRUE)
     }, add = TRUE)
   }
 
@@ -310,7 +442,7 @@ package_coverage <- function(path = ".",
   # install the package in a temporary directory
   withr::with_makevars(flags, assignment = "+=",
     utils::install.packages(repos = NULL,
-                            lib = tmp_lib,
+                            lib = install_path,
                             pkg$path,
                             type = "source",
                             INSTALL_opts = c("--example",
@@ -322,39 +454,45 @@ package_coverage <- function(path = ".",
                             quiet = quiet))
 
   # add hooks to the package startup
-  add_hooks(pkg$package, tmp_lib,
+  add_hooks(pkg$package, install_path,
     fix_mcexit = should_enable_parallel_mcexit_fix(pkg))
 
-  libs <- env_path(tmp_lib, .libPaths())
+  libs <- env_path(install_path, .libPaths())
+
+  # We need to set the libpaths in the current R session for examples with
+  # install or runtime Sexpr blocks, which may implicitly load the package in
+  # the current R session.
+  withr::with_libpaths(install_path, action = "prefix", {
 
   withr::with_envvar(
     c(R_DEFAULT_PACKAGES = "datasets,utils,grDevices,graphics,stats,methods",
       R_LIBS = libs,
       R_LIBS_USER = libs,
       R_LIBS_SITE = libs,
-      R_COVR = "true"), {
+      R_COVR = "true",
+      R_TESTS = file.path(R.home("share"), "R", "tests-startup.R")), {
 
 
     withCallingHandlers({
       if ("vignettes" %in% type) {
         type <- type[type != "vignettes"]
-        run_vignettes(pkg, tmp_lib)
+        run_vignettes(pkg, install_path)
       }
 
-      out_dir <- file.path(tmp_lib, pkg$package)
+      out_dir <- file.path(install_path, pkg$package)
       if ("examples" %in% type) {
         type <- type[type != "examples"]
         # testInstalledPackage explicitly sets R_LIBS="" on windows, and does
         # not restore it after, so we need to reset it ourselves.
         withr::with_envvar(c(R_LIBS = Sys.getenv("R_LIBS")), {
-          result <- tools::testInstalledPackage(pkg$package, outDir = out_dir, types = "examples", lib.loc = tmp_lib, ...)
+          result <- tools::testInstalledPackage(pkg$package, outDir = out_dir, types = "examples", lib.loc = install_path, ...)
           if (result != 0L) {
             show_failures(out_dir)
           }
         })
       }
       if ("tests" %in% type) {
-        result <- tools::testInstalledPackage(pkg$package, outDir = out_dir, types = "tests", lib.loc = tmp_lib, ...)
+        result <- tools::testInstalledPackage(pkg$package, outDir = out_dir, types = "tests", lib.loc = install_path, ...)
         if (result != 0L) {
           show_failures(out_dir)
         }
@@ -362,14 +500,15 @@ package_coverage <- function(path = ".",
 
       # We always run the commands file (even if empty) to load the package and
       # initialize all the counters to 0.
-      run_commands(pkg, tmp_lib, code)
+      run_commands(pkg, install_path, code)
     },
     message = function(e) if (quiet) invokeRestart("muffleMessage") else e,
     warning = function(e) if (quiet) invokeRestart("muffleWarning") else e)
     })
+    })
 
   # read tracing files
-  trace_files <- list.files(path = tmp_lib, pattern = "^covr_trace_[^/]+$", full.names = TRUE)
+  trace_files <- list.files(path = install_path, pattern = "^covr_trace_[^/]+$", full.names = TRUE)
   coverage <- merge_coverage(trace_files)
   if (!uses_icc()) {
     res <- run_gcov(pkg$path, quiet = quiet, clean = clean)
@@ -377,13 +516,14 @@ package_coverage <- function(path = ".",
     res <- run_icov(pkg$path, quiet = quiet)
   }
 
-  coverage <- structure(c(coverage, res),
-      class = "coverage",
-      package = pkg,
-      relative = relative_path)
+  coverage <- as_coverage(
+    c(coverage, res),
+    package = pkg,
+    root = root
+  )
 
   if (!clean) {
-    attr(coverage, "library") <- tmp_lib
+    attr(coverage, "library") <- install_path
   }
 
   if (getOption("covr.filter_non_package", TRUE)) {
@@ -403,7 +543,7 @@ package_coverage <- function(path = ".",
   exclude(coverage,
     line_exclusions = line_exclusions,
     function_exclusions = function_exclusions,
-    path = if (isTRUE(relative_path)) pkg$path else NULL)
+    path = root)
 }
 
 #' Convert a coverage dataset to a list
@@ -449,34 +589,109 @@ show_failures <- function(dir) {
 # merge multiple coverage files together. Assumes the order of coverage lines
 # is the same in each object, this should always be the case if the objects are
 # from the same initial library.
-merge_coverage <- function(files) {
-  nfiles <- length(files)
-  if (nfiles == 0) {
+merge_coverage <- function(x) {
+  UseMethod("merge_coverage")
+}
+
+merge_coverage.character <- function(files) {
+  coverage_objs <- lapply(files, function(f) {
+    as.list(suppressWarnings(readRDS(f)))
+  })
+  merge_coverage(coverage_objs)
+}
+
+merge_coverage.list <- function(coverage_objs) {
+  if (length(coverage_objs) == 0) {
     return()
   }
 
-  x <- suppressWarnings(readRDS(files[1]))
-  x <- as.list(x)
-  if (nfiles == 1) {
-    return(x)
-  }
-
+  x <- coverage_objs[[1]]
   names <- names(x)
-  for (i in 2:nfiles) {
-    y <- suppressWarnings(readRDS(files[i]))
+  clean_coverage_tests(x)  # x[[key]]$tests environments modified in-place
+
+  for (y in tail(coverage_objs, -1L)) {
+
+    # only affects coverage produced with options(covr.record_tests = TRUE)
+    clean_coverage_tests(y)
+    x <- merge_coverage_tests(from = y, into = x)
+
     for (name in intersect(names, names(y))) {
+      if (name == "tests") next
       x[[name]]$value <- x[[name]]$value + y[[name]]$value
     }
+
     for (name in setdiff(names(y), names)) {
       x[[name]] <- y[[name]]
     }
+
     names <- union(names, names(y))
-    y <- NULL
   }
 
   x
 }
 
+# Strip allocated, but unused test records from coverage test matrix
+#
+# The tally of tests that hit each trace is held in a pre-allocated matrix
+# which may be padded with unused rows. Start by stripping unused rows:
+#
+# If tests were not recorded (that is, if `options(covr.record_tests)` was not
+# `TRUE` when the coverage was calculated, this function will have no effect.
+#
+# @param obj A coverage counter environment, within which a $tests$tally matrix
+#   may have been allocated, but not entirely populated.
+#
+clean_coverage_tests <- function(obj) {
+  counter_has_tests_tally <- function(counter) !is.null(counter$tests)
+  if (is.na(Position(counter_has_tests_tally, obj))) return()
+
+  for (i in seq_along(obj)) {
+    if (is.null(val <- obj[[i]]$value)) next
+    if (is.null(n <- nrow(obj[[i]]$tests$tally)) || n < val) next
+    obj[[i]]$tests$tally <- obj[[i]]$tests$tally[seq_len(val),,drop = FALSE]
+  }
+}
+
+# Merge recorded tests from one coverage object into another. Because coverage
+# objects are environments, these environments will be modified by-reference as
+# a side-effect of calling this function.
+#
+# If tests were not recorded (that is, if `options(covr.record_tests)` was not
+# `TRUE` when the coverage was calculated, this function will have no effect.
+#
+# @param from A coverage counter environment whose tests should be merged into
+#   \code{into}
+# @param into A coverage counter environment to add tests into
+#
+merge_coverage_tests <- function(from, into = NULL) {
+  if (is.null(from$tests)) return(into)
+
+  # TODO: The x[[name]]$tests$tally matrices are re-allocated with each rbind of
+  # additional test hits as each object is merged. This could be avoided by
+  # first calculating the total rows needed to store all the merged tests and
+  # then allocating a matrix of the appropriate size from the start. In most
+  # cases, this amounts to neglegable overhead but is an opportunity for
+  # improvement.
+
+  # align tests from coverage objects
+  test_idx <- match(names(from$tests), Filter(nchar, names(into$tests)))
+  new_test_idx <- if (!length(test_idx)) seq_along(from$tests) else which(is.na(test_idx))
+  test_idx[new_test_idx] <- length(into$tests) + seq_along(new_test_idx)
+
+  # append any tests that we haven't encountered in previous objects
+  into$tests <- append(into$tests, from$tests[new_test_idx])
+  from$tests <- NULL
+
+  # modify trace test tallies
+  for (name in intersect(names(into), names(from))) {
+    if (name == "tests") next
+    from[[name]]$tests$tally[,1L] <- test_idx[from[[name]]$tests$tally[,1L]]
+    into[[name]]$tests$tally <- rbind(into[[name]]$tests$tally, from[[name]]$tests$tally)
+  }
+
+  into
+}
+
 parse_type <- function(type) {
   type <- match_arg(type, choices = c("tests", "vignettes", "examples", "all", "none"), several.ok = TRUE)
   if (type %==% "all") {
@@ -519,9 +734,9 @@ run_vignettes <- function(pkg, lib) {
 run_commands <- function(pkg, lib, commands) {
   outfile <- file.path(lib, paste0(pkg$package, "-commands.Rout"))
   failfile <- paste(outfile, "fail", sep = "." )
-  cat(
-    "library('", pkg$package, "')\n",
-    commands, "\n", file = outfile, sep = "")
+  writeLines(c(
+    paste0("library('", pkg$package, "')"),
+    commands), con = outfile)
   cmd <- paste(shQuote(file.path(R.home("bin"), "R")),
                "CMD BATCH --vanilla --no-timing",
                shQuote(outfile), shQuote(failfile))
@@ -545,13 +760,16 @@ run_commands <- function(pkg, lib, commands) {
 # @param pkg_name name of the package to add hooks to
 # @param lib the library path to look in
 # @param fix_mcexit whether to add the fix for mcparallel:::mcexit
-add_hooks <- function(pkg_name, lib, fix_mcexit = FALSE) {
+add_hooks <- function(pkg_name, lib, fix_mcexit = FALSE,
+  record_tests = isTRUE(getOption("covr.record_tests", FALSE))) {
+
   trace_dir <- paste0("Sys.getenv(\"COVERAGE_DIR\", \"", lib, "\")")
 
   load_script <- file.path(lib, pkg_name, "R", pkg_name)
   lines <- readLines(file.path(lib, pkg_name, "R", pkg_name))
   lines <- append(lines,
-    c("setHook(packageEvent(pkg, \"onLoad\"), function(...) covr:::trace_environment(ns))",
+    c(paste0("setHook(packageEvent(pkg, \"onLoad\"), function(...) options(covr.record_tests = ", record_tests, "))"),
+      "setHook(packageEvent(pkg, \"onLoad\"), function(...) covr:::trace_environment(ns))",
       paste0("reg.finalizer(ns, function(...) { covr:::save_trace(", trace_dir, ") }, onexit = TRUE)")),
     length(lines) - 1L)
 
